/*------------------------------------------------*/
/* 07osvetleni/main.c                             */
#define _CRT_SECURE_NO_WARNINGS
#define GLEW_STATIC
#include <GL/glew.h>
#include <stdio.h>
#include <GLFW/glfw3.h>
#include <stdbool.h>
#include <math.h>
#include "linmath-extended.h"
#include "vertex-data.h"
#include "shader.h"
#include "normals.h"
#include "update-angles.h"
#include "events.h"

void initVariables(Shader *program);
void initBuffers(Shader *program);
void uninitBuffers();
void ticking(Shader * program, GLFWwindow* window);
void draw(Shader *program, GLFWwindow* window, GLuint *VAO);
void animate();

GLuint VAO[2], BO[7];
struct {
	GLuint uPMatrixLoc;
	GLuint uMVMatrixLoc;
	GLuint uNMatrixLoc;
	GLuint uAmbientColorLoc;
	GLuint uUseDirectionalLoc;
	GLuint uDirectionalColorLoc;
	GLuint uLightingPositionLoc;
	GLuint uUseSpecularLoc;
} variables;
mat4x4 rotationMatrix;
float x, y, z; // light position

int main(void)
{
	GLFWwindow* window;
	/* Initialize the library */
	if (!glfwInit()) {
		return -1;
	}
	glfwWindowHint(GLFW_CONTEXT_VERSION_MAJOR, 3);
	glfwWindowHint(GLFW_CONTEXT_VERSION_MINOR, 3);
	glfwWindowHint(GLFW_OPENGL_PROFILE, GLFW_OPENGL_CORE_PROFILE);

	/* Create a windowed mode window and its OpenGL context */
	window = glfwCreateWindow(640, 480, "07 - Osvetleni", NULL, NULL);
	if (!window)
	{
		printf("Nelze inicializovat kontext\n");
		glfwTerminate();
		return -1;
	}
	/* Make the window's context current */
	glfwMakeContextCurrent(window);
	printf("%s\n", glGetString(GL_VERSION));

	/* Load modern OpenGL functions */
	glewExperimental = GL_TRUE;
	glewInit();

	/* events */
	glfwSetKeyCallback(window, key_callback);
	glfwSetMouseButtonCallback(window, mouse_button_callback);
	glfwSetCursorPosCallback(window, cursor_position_callback);

	//glfwSwapInterval(1);

        Shader program = shader("default.vs", "default.fs");
	if(!program.Program) {
		return -1;
	}

	initVariables(&program);
	initBuffers(&program);

	// setup GL state
	glEnable(GL_DEPTH_TEST);
	mat4x4_identity(rotationMatrix);
	mat4x4_rotate_Y(rotationMatrix, rotationMatrix, -3.14159/14.0);
	mat4x4_rotate_X(rotationMatrix, rotationMatrix, 3.14159/2.5);
	useColor.ambient = 1;
	useColor.directional = 1;
	useColor.specular = 1;

	ticking(&program, window);

	uninitBuffers();

	glfwDestroyWindow(window);
	glfwTerminate();
	return 0;
}

void initVariables(Shader *program)
{
	variables.uMVMatrixLoc = program->getUniformLocation(program, "uMVMatrix");
	variables.uPMatrixLoc = program->getUniformLocation(program, "uPMatrix");
	variables.uNMatrixLoc = program->getUniformLocation(program, "uNMatrix");
	variables.uAmbientColorLoc = program->getUniformLocation(program, "uAmbientColor");
	variables.uUseDirectionalLoc = program->getUniformLocation(program, "uUseDirectional");
	variables.uDirectionalColorLoc = program->getUniformLocation(program, "uDirectionalColor");
	variables.uLightingPositionLoc = program->getUniformLocation(program, "uLightingPosition");

	variables.uUseSpecularLoc = program->getUniformLocation(program, "uUseSpecular");
}

void initBuffers(Shader *program)
{
        glGenVertexArrays(sizeof(VAO)/sizeof(VAO[0]), VAO);
        glGenBuffers(sizeof(BO)/sizeof(BO[0]), BO);

	glBindVertexArray(VAO[0]);
	{
		glBindBuffer(GL_ARRAY_BUFFER, BO[0]);
	        glBufferData(GL_ARRAY_BUFFER, sizeof(vertices), vertices, GL_STATIC_DRAW);
		glVertexAttribPointer(0, 3, GL_FLOAT, GL_FALSE, 3 * sizeof(GLfloat), (GLvoid*)0);
	        glEnableVertexAttribArray(0);
	
		glBindBuffer(GL_ARRAY_BUFFER, BO[1]);
	        glBufferData(GL_ARRAY_BUFFER, sizeof(colors), colors, GL_STATIC_DRAW);
		glVertexAttribPointer(1, 4, GL_FLOAT, GL_FALSE, 4 * sizeof(GLfloat), (GLvoid*)0);
	        glEnableVertexAttribArray(1);

		const int directions[] = {1,1,1,1,1,1};
		ComputedNormals normals = createNormals(vertices, sizeof(vertices)/sizeof(vertices[0]), directions);
		glBindBuffer(GL_ARRAY_BUFFER, BO[5]);
	        glBufferData(GL_ARRAY_BUFFER, sizeof(GLfloat)*normals.length, normals.normals, GL_STATIC_DRAW);
		glVertexAttribPointer(2, 3, GL_FLOAT, GL_FALSE, 3 * sizeof(GLfloat), (GLvoid*)0);
	        glEnableVertexAttribArray(2);
		normals.free(&normals);
	}

	glBindVertexArray(VAO[1]);
	{
		glBindBuffer(GL_ARRAY_BUFFER, BO[2]);
	        glBufferData(GL_ARRAY_BUFFER, sizeof(lines), lines, GL_STATIC_DRAW);
		glVertexAttribPointer(0, 3, GL_FLOAT, GL_FALSE, 3 * sizeof(GLfloat), (GLvoid*)0);
	        glEnableVertexAttribArray(0);

		glBindBuffer(GL_ARRAY_BUFFER, BO[3]);
	        glBufferData(GL_ARRAY_BUFFER, sizeof(lineColors), lineColors, GL_STATIC_DRAW);
		glVertexAttribPointer(1, 4, GL_FLOAT, GL_FALSE, 4 * sizeof(GLfloat), (GLvoid*)0);
	        glEnableVertexAttribArray(1);

		ComputedNormals normals = createNormalsByIndex(
					lines, sizeof(lines)/sizeof(lines[0]),
					lineIndices, sizeof(lineIndices)/sizeof(lineIndices[0]),
					NULL
				);
		glBindBuffer(GL_ARRAY_BUFFER, BO[6]);
	        glBufferData(GL_ARRAY_BUFFER, sizeof(GLfloat)*normals.length, normals.normals, GL_STATIC_DRAW);
		glVertexAttribPointer(2, 3, GL_FLOAT, GL_FALSE, 3 * sizeof(GLfloat), (GLvoid*)0);
	        glEnableVertexAttribArray(2);
		normals.free(&normals);

		glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, BO[4]);
	        glBufferData(GL_ELEMENT_ARRAY_BUFFER, sizeof(lineIndices), lineIndices, GL_STATIC_DRAW);
	}
	glBindVertexArray(0); // Unbind VAO
	glBindBuffer(GL_ARRAY_BUFFER, 0);
}

void uninitBuffers()
{
	glDeleteVertexArrays(sizeof(VAO)/sizeof(VAO[0]), VAO);
	glDeleteBuffers(sizeof(BO)/sizeof(BO[0]), BO);
}

void ticking(Shader * program, GLFWwindow* window) {
	/* Loop until the user closes the window */
	while (!glfwWindowShouldClose(window))
	{
		animate();
		draw(program, window, VAO);
		glfwSwapBuffers(window);
		glfwPollEvents();
	}
}

void animate() {
	float rLight = glfwGetTime()/1.0;
	x = 2.0*cos(rLight);
	y = 2.0*sin(rLight);
	z = -9.5;
}

void draw(Shader *program, GLFWwindow* window, GLuint *VAO)
{
       	int width, height;
        glfwGetFramebufferSize(window, &width, &height);
        glViewport(0, 0, width, height);

	glClearColor(0.f, 0.f, 0.f, 1.f);
	glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);

       	program->Use(program);

	mat4x4 uMVMatrix, uPMatrix;
	mat3x3 uNMatrix;

	mat4x4_perspective(uPMatrix, degToRad(45.0), ((GLfloat) width)/height, 1.0, 100.0);
	mat4x4_translate(uMVMatrix, 0.0, 0.0, -10.0);
	mat4x4_mul(uMVMatrix, uMVMatrix, rotationMatrix);

	/*
	 * http://learningwebgl.com/blog/?p=684
	 * Normal = mat3(transpose(inverse(model))) * normal;
	 */
	mat3x3_from_mat4x4(uNMatrix, uMVMatrix);
	mat3x3_invert(uNMatrix, uNMatrix);
	mat3x3_transpose(uNMatrix, uNMatrix);

        glUniformMatrix4fv(variables.uMVMatrixLoc, 1, GL_FALSE, (const GLfloat *) uMVMatrix);
        glUniformMatrix4fv(variables.uPMatrixLoc, 1, GL_FALSE, (const GLfloat *) uPMatrix);
        glUniformMatrix3fv(variables.uNMatrixLoc, 1, GL_FALSE, (const GLfloat *) uNMatrix);

	if(useColor.ambient) {
	        glUniform3f(variables.uAmbientColorLoc, 0.2, 0.2, 0.2);
	} else {
	        glUniform3f(variables.uAmbientColorLoc, 0.0, 0.0, 0.0);
	}
        glUniform1i(variables.uUseDirectionalLoc, useColor.directional);
	glUniform3f(variables.uDirectionalColorLoc, 0.8, 0.8, 0.8);
	glUniform3f(variables.uLightingPositionLoc, x, y, z);
        glUniform1i(variables.uUseSpecularLoc, useColor.specular);

	// Vykresli osy
        glBindVertexArray(VAO[1]);
	glDrawElements(GL_TRIANGLES, sizeof(lineIndices)/sizeof(lineIndices[0]), GL_UNSIGNED_INT, 0);
        glBindVertexArray(0);

	// Vykresli jehlan 
        glBindVertexArray(VAO[0]);
        glDrawArrays(GL_TRIANGLES, 0, sizeof(vertices)/sizeof(vertices[0])/3);
        glBindVertexArray(0);
}

/*------------------------------------------------*/
