#include "model.h"
#include "mesh.h"

static void drawMesh(Mesh *mesh, Model *model);

Mesh constructMesh(void) {
	Mesh mesh;
	mesh.hasTextures = false;
	mesh.vertices = NULL;
	mesh.indices = NULL;
	mesh.textures = NULL;
	mesh.Draw = drawMesh;
	return mesh;
}

void setupMesh(Mesh * mesh) {
	glGenVertexArrays(1, &mesh->VAO);
	glGenBuffers(1, &mesh->VBO);
	glGenBuffers(1, &mesh->EBO);

	glBindVertexArray(mesh->VAO);
	glBindBuffer(GL_ARRAY_BUFFER, mesh->VBO);
	glBufferData(GL_ARRAY_BUFFER, mesh->numVertices * sizeof(Vertex), &mesh->vertices[0], GL_STATIC_DRAW);

	// Vertex Positions
	glEnableVertexAttribArray(0);
	glVertexAttribPointer(0, 3, GL_FLOAT, GL_FALSE, sizeof(Vertex), (GLvoid*)0);
	// Vertex Normals
	glEnableVertexAttribArray(1);
	glVertexAttribPointer(1, 3, GL_FLOAT, GL_FALSE, sizeof(Vertex), (GLvoid*)offsetof(Vertex, Normal));
	// Vertex Texture Coords
	glEnableVertexAttribArray(2);
	glVertexAttribPointer(2, 2, GL_FLOAT, GL_FALSE, sizeof(Vertex), (GLvoid*)offsetof(Vertex, TexCoords));

	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, mesh->EBO);
	glBufferData(GL_ELEMENT_ARRAY_BUFFER, mesh->numIndices * sizeof(GLuint), &mesh->indices[0], GL_STATIC_DRAW);

	glBindVertexArray(0);
}


void drawMesh(Mesh *mesh, Model *model)
{
	size_t i = 0;
	Shader *program = &(model->program);

	GLuint texIx = 0;
	for (i = 0; i < mesh->numTextures; i++)
	{
		switch (mesh->textures[i].type)
		{
		case aiTextureType_AMBIENT:
			glUniform1i(program->getUniformLocation(program, "texture_ambient1"), texIx);
			break;
		case aiTextureType_DIFFUSE:
			glUniform1i(program->getUniformLocation(program, "texture_diffuse1"), texIx);
			break;
		case aiTextureType_SPECULAR:
			glUniform1i(program->getUniformLocation(program, "texture_specular1"), texIx);
			break;
		default:
			continue;
		}
		glActiveTexture(GL_TEXTURE0 + texIx);
		glBindTexture(GL_TEXTURE_2D, mesh->textures[i].id);
		texIx++;
	}

	glUniform4fv(program->getUniformLocation(program, "ambientColor"), 1, mesh->ambientColor);
	glUniform4fv(program->getUniformLocation(program, "diffuseColor"), 1, mesh->diffuseColor);
	glUniform4fv(program->getUniformLocation(program, "specularColor"), 1, mesh->specularColor);
	glUniform4fv(program->getUniformLocation(program, "emissiveColor"), 1, mesh->emissiveColor);
	glUniform1f(program->getUniformLocation(program, "shininess"), mesh->shininess);
	glUniform1i(program->getUniformLocation(program, "ambientTexCount"), mesh->ambientTexCount);
	glUniform1i(program->getUniformLocation(program, "diffuseTexCount"), mesh->diffuseTexCount);
	glUniform1i(program->getUniformLocation(program, "specularTexCount"), mesh->specularTexCount);

	glBindVertexArray(mesh->VAO);
	glDrawElements(GL_TRIANGLES, mesh->numIndices, GL_UNSIGNED_INT, 0);
	glBindVertexArray(0);
	/*
	for (i = 0; i < mesh->numTextures; i++)
	{
		glActiveTexture(GL_TEXTURE0 + i);
		glBindTexture(GL_TEXTURE_2D, 0);
	}*/
}
